    module SPTpol_bb100d
    use cmbtypes
    use likelihood
    use MatrixUtils
    use settings

    implicit none


    ! Bandpowers are read in this order (all BB):
    !            (90x90), (90x150), (150x150)
    integer :: nfreq, nband, nbin, nall, bands_per_freq
    double precision, dimension(:,:), allocatable :: spec, cov
    double precision, dimension(:,:), allocatable :: windows
    integer :: spt_windows_lmin, spt_windows_lmax
    double precision, dimension(:), allocatable :: cl_to_dl_conversion, dl_th, poisson_constant
    double precision, dimension(:), allocatable :: dust_template, tensor_spectrum, scalar_spectrum
    
    logical :: SuccessfulSPTInitialization
    logical :: use_dust_gaussian_prior
    
    type, extends(CosmologyLikelihood) :: SPTpolLikelihood
     contains
       procedure :: LogLike => SPTpolLnLike
    end type SPTpolLikelihood

    contains

    subroutine SPTpolLikelihood_Add(LikeList, Ini)
      use IniFile
      use settings
      class(LikelihoodList) :: LikeList
      Type(TIniFile) :: ini
      Type(SPTpolLikelihood), pointer :: like

      character (LEN=Ini_max_string_len) :: desc_file
      character (LEN=Ini_max_string_len) :: bp_file
      character (LEN=Ini_max_string_len) :: cov_file
      character (LEN=Ini_max_string_len) :: dust_file
      character (LEN=Ini_max_string_len) :: tensor_file
      character (LEN=Ini_max_string_len) :: scalar_file
      character (LEN=Ini_max_string_len) :: win_folder
      

      if (.not. Ini_Read_Logical_File(Ini, 'use_SPTpol', .false.)) then
         return
      end if

      !Are we placing a Gaussian prior on dust?
      use_dust_gaussian_prior = Ini_Read_logical_File(Ini, 'use_dust_gaussian_prior', .false.)

      allocate(like)
      Like%speed = 10
      Like%LikelihoodType = 'SPTpol'
      Like%name = 'SPTpol_bb100d'
      Like%needs_background_functions = .true.

      print *, 'Likelihood Name: ', Like%name

      call Like%loadParamNames(trim(DataDir)//'SPTpol.paramnames')
      call LikeList%Add(like)

      !I could simplify this by moving all the relevant info from
      !desc_file into the ini file.
      desc_file = Ini_Read_String_Default_File(Ini, 'sptpol_desc_file', &
           trim(DataDir)//'bb100d.desc_file')

      !Read in the bandpower and cov files, and determine where the windows are.
      bp_file = Ini_Read_String_Default_File(Ini, 'sptpol_bp_file','')
      cov_file = Ini_Read_String_Default_File(Ini, 'sptpol_cov_file','')
      dust_file = Ini_Read_String_Default_File(Ini, 'sptpol_dust_file','')
      tensor_file = Ini_Read_String_Default_File(Ini, 'sptpol_tensor_file','')
      scalar_file = Ini_Read_String_Default_File(Ini, 'sptpol_scalar_file','')
      win_folder = Ini_Read_String_Default_File(Ini, 'sptpol_window_dir','')
      if (bp_file=='' .or. desc_file=='' .or. cov_file=='' .or. dust_file=='' .or. tensor_file=='' .or. scalar_file=='' .or. win_folder=='' ) then
         print*,'Missing required sptpol key: received: ',bp_file,desc_file,cov_file,dust_file,tensor_file,scalar_file,win_folder
         stop
      endif

      call InitSPTpolData(desc_file, bp_file, cov_file, dust_file, tensor_file, scalar_file, win_folder)
    end subroutine SPTpolLikelihood_Add

    subroutine InitSPTpolData(desc_file, bp_file, cov_file, dust_file, tensor_file, scalar_file, win_folder)
      use IniFile
      character(LEN=Ini_max_string_len) :: desc_file, bp_file, cov_file, dust_file, tensor_file, scalar_file
      character(LEN=Ini_max_string_len) :: win_folder
      integer i,j,k,dum,lwin
      real rtmp

      !Obtain necessary info from the desc_file pertaining
      !to which freqs we're using, ell range, and number of bins per spectrum.
      call OpenTxtFile(desc_file, tmp_file_unit)
      print *, 'Reading desc file...',trim(desc_file)

      read (tmp_file_unit,*) nbin,nfreq !Number of bandpowers per spectrum, Number of freqs
      read (tmp_file_unit,*) spt_windows_lmin, spt_windows_lmax !Min and Max ell in window files.
      close(tmp_file_unit)

      print *, 'nbin: ', nbin
      print *, 'nfreq: ', nfreq
      print *, 'spt_windows_lmin: ', spt_windows_lmin
      print *, 'spt_windows_lmax: ', spt_windows_lmax
      print *, 'win_folder: ', trim(win_folder)

      !We have only BB for three cross-frequencies
      bands_per_freq = 1
     
      nband = bands_per_freq*(nfreq)
      nall = nbin*(nfreq*bands_per_freq)

      print *, 'nband: ', nband
      print *, 'nall: ', nall
      print *, 'bands_per_freq: ', bands_per_freq

      if (lmax .lt. spt_windows_lmax) then
         spt_windows_lmax=lmax
      end if

      if (spt_windows_lmin < 2 .or. spt_windows_lmin >= spt_windows_lmax) then
         call mpistop('Invalid lranges for sptpol')
      end if
      
      if (spt_windows_lmin .lt. 2) then
         print *, "Unexpected lmin for windows!"
         stop
      end if

      allocate(windows(spt_windows_lmin:spt_windows_lmax,1:nall), &
           spec(1:nbin,nband))

      allocate(cov(1:nall,1:nall))
      
      allocate(poisson_constant(spt_windows_lmin:spt_windows_lmax))
      do j=spt_windows_lmin,spt_windows_lmax
         poisson_constant(j) = 1.0d0
      end do

      allocate(dust_template(spt_windows_lmin:spt_windows_lmax), &
           tensor_spectrum(spt_windows_lmin:spt_windows_lmax), &
           scalar_spectrum(spt_windows_lmin:spt_windows_lmax), &
           dl_th(spt_windows_lmin:spt_windows_lmax))

      !Read in bandpowers
      !Should be 90x90, 90x150, 150x150 BB in that order from SPTpol analysis.
      print *, 'Reading in bandpowers...'
      print *, 'Reading bp_file...',trim(bp_file)
      call OpenTxtFile(bp_file, tmp_file_unit)
      do i=1,nband
         do j=1,nbin
            read (tmp_file_unit,*) dum,spec(j,i)
         end do
      end do
      close(tmp_file_unit)
      print *, 'Bandpowers are:', spec
     
      !Read in dust template (150 GHz).
      print *, 'Reading in dust template...'
      print *, 'Reading dust file...',trim(dust_file)
      call OpenTxtFile(dust_file, tmp_file_unit)
      do j=spt_windows_lmin,spt_windows_lmax
         read (tmp_file_unit,*) dum,dust_template(j)
      end do
      close(tmp_file_unit)

      !Read in pre-calculated tensor BB.
      print *, 'Reading in tensor spectrum template...'
      call OpenTxtFile(tensor_file, tmp_file_unit)
      do j=spt_windows_lmin,spt_windows_lmax
         read (tmp_file_unit,*) dum, tensor_spectrum(j)
      end do
      close(tmp_file_unit)

      !Read in pre-calculated scalar BB.
      print *, 'Reading in scalar spectrum template...'
      call OpenTxtFile(scalar_file, tmp_file_unit)
      do j=spt_windows_lmin,spt_windows_lmax
         read (tmp_file_unit,*) dum, scalar_spectrum(j)
      end do
      close(tmp_file_unit)
    

      !Read in covariance
      !Should be 90x90, 90x150, 150x150 BB
      print *, 'Reading in cov_data matrix...'
      call OpenTxtFile(cov_file, tmp_file_unit)
      do i=1,nall
         do j=1,nall
            read (tmp_file_unit,*) cov(i,j)
         end do
      end do
      close (tmp_file_unit)

      call Matrix_CholeskyDouble(cov)

      ! Read in windows
      !Should be 90x90, 90x150, 150x150
      print *, 'Reading in windows...'
      do i=1,nall
         call OpenTxtFile(trim(win_folder)//trim(numcat('/window_',i)),tmp_file_unit)
         do j=spt_windows_lmin,spt_windows_lmax
            read (tmp_file_unit,*) dum, windows(j,i)
         end do
         close(tmp_file_unit)
      end do

      SuccessfulSPTInitialization = .true.
    end subroutine InitSPTpolData

    function SPTpolLnLike(like, CMB, Theory, DataParams)
      class(SPTpolLikelihood) :: like
      class(CMBParams) :: CMB
      class(TheoryPredictions) :: Theory !cl(lmax, num_tot_cls) TT, TE, EE, BB,(+phi) in that order from Camb.
      real(mcp) :: SPTpolLnLike
      real(mcp) :: A_dustLnLike
      real(mcp) :: dum
      double precision :: DataParams(:) !Foreground parameters (plus wimp annihilation vector amplitude)
      double precision, parameter :: sigmaA_dust = 0.30 
      double precision, parameter :: meanA_dust = 1.0 ! Scales all cross-frequencies.
      double precision, dimension(1:nall) :: deltacb
      double precision, dimension(1:nbin) :: tmpcb
      double precision, dimension(1) :: junk
      integer :: j,k,l
      real(mcp) :: norm

      do j=1,(nfreq)      
         tmpcb(:)=0
          
         !DO EVERYTHING IN CL.
         
         !DataParams: [A_lens, A_tens, A_dust, A_PS90, A_PS90150, A_PS150]
         !First get model CMB spectra (in Cl).
         dl_th(:) = DataParams(1)*scalar_spectrum(:) + DataParams(2)*tensor_spectrum(:)
         
         !Now add foreground terms.
         !Case 1: 90x90
         if (j==1) then
             dl_th(:) = dl_th(:) + DataParams(3)*0.00169*dust_template(:) + DataParams(4)*4.65*10**(-8.0d0)*poisson_constant(:) 
         end if 

         !Case 2: 90x150
         if (j==2) then
             dl_th(:) = dl_th(:) + DataParams(3)*0.00447*dust_template(:) + DataParams(5)*2.27*10**(-8.0d0)*poisson_constant(:) 
         end if 

         !Case 2: 150x150
         if (j==3) then
             dl_th(:) = dl_th(:) + DataParams(3)*0.0118*dust_template(:) + DataParams(6)*1.11*10**(-8.0d0)*poisson_constant(:) 
         end if 
          
         !Now bin into bandpowers with the window functions - output is tmpcb (binned theory+foreground spectrum).
         call dgemv('T',spt_windows_lmax-spt_windows_lmin+1,nbin,1.0d0,&
              windows(:,1+nbin*(j-1):nbin+nbin*(j-1)),spt_windows_lmax-spt_windows_lmin+1,&
              dl_th,1,0d0,tmpcb,1)

         !Finally, get \Delta Cl's between our model Cls and our input spectra
         !for covariance calculation.
         !spec(:,1) = 90x90
         !spec(:,2) = 90x150
         !spec(:,3) = 150x150
         deltacb(1+(j-1)*nbin:nbin+(j-1)*nbin) = tmpcb(:) - spec(:,j)
      end do

      !Now calculate the SPTpol likelihood.
      !SPTpolLnLike = Matrix_GaussianLogLikeCholDouble(cov,deltacb) 

      !Here, the covariance doesn't change with parameter values, so we can just calculate chi^2 for
      !the likelihood.
      SPTpolLnLike = Matrix_GaussianHalfChisqCholDouble(cov,deltacb) 

      !Calculate the Gaussian prior term for the dust, ignoring the extra log|cov| term.
      if (use_dust_gaussian_prior== .true.) then
         A_dustLnLike = 0.5*((DataParams(3) - meanA_dust)**2.0d0 / sigmaA_dust**2.0d0)
      else
         A_dustLnLike = 0.0d0
      end if

      print *, 'SPTpolLnLike = ', SPTpolLnLike
      print *, 'A_dustLnLike = ', A_dustLnLike

      SPTpolLnLike = SPTpolLnLike + A_dustLnLike

      if (feedback > 1)      print *, 'SPTpolLnLike = ', SPTpolLnLike
    end function SPTpolLnLike

    end module SPTpol_bb100d
